︠385e2885-1cad-4fdf-8786-f5c6716e4a08asi︠
%auto

def plot_phase_portrait(field, x_range, y_range, init_conds, **options):
    """Plot a vector field, together with equilibrium points and trajectories

    'x_range' should have the form (x, xmin, xmax), and likewise for 'y_range'.
    'init_conds' should be a dictionary whose keys are the points that are to
        be used as initial conditions for trajectories. For each such initial
        condition, the value in the dictionary should be the range of t values
        over which to plot the trajectory starting at that initial point, such
        as srange(0, 20, 0.01) for t from 0 to 20 with a step size of 0.01. If
        you don't want to plot any trajectories, just use {} for this.

    Options:
        'field_color' - The color for vector field arrows (default "limegreen")
        'nullcline_colors' - A pair of colors for the two nullclines (default
            ("lightskyblue", "lightsalmon") )
        'eqpt_color' - The color for equilibrium points (default "darkmagenta")
        'trajectory_color' - The color for trajectories (default "red")
        Specifying None (or any False value) for any of the above will cause
        it to not be plotted at all. (Note that for the nullclines, this must
        be a tuple (None, None).)
    All other options are as for the various 'plot*' commands, and are passed
    directly on to them.
    """

    # Initialization
    x, xmin, xmax = x_range
    y, ymin, ymax = y_range
    field_color = options.pop("field_color", "limegreen")
    xnullcline_color, ynullcline_color = options.pop("nullcline_colors",
            ("lightskyblue", "lightsalmon"))
    eqpt_color = options.pop("eqpt_color", "darkmagenta")
    trajectory_color = options.pop("trajectory_color", "red")
    options.setdefault("frame", False)
    options.setdefault("axes", True)
    options.setdefault("aspect_ratio", "auto")
    options.setdefault("xmin", xmin)
    options.setdefault("xmax", xmax)
    options.setdefault("ymin", ymin)
    options.setdefault("ymax", ymax)
    p = Graphics()

    # Plot the vector field itself
    if field_color:
        options["color"] = field_color
        p += plot_vector_field(field, x_range, y_range, **options)

    # Plot the "vertical" nullcline
    if xnullcline_color:
        options["color"] = xnullcline_color
        p += implicit_plot(field[0] == 0, x_range, y_range, linewidth=2, **options)

    # Plot the "horizontal" nullcline
    if ynullcline_color:
        options["color"] = ynullcline_color
        p += implicit_plot(field[1] == 0, x_range, y_range, linewidth=2, **options)

    # Find and plot the equilibrium points
    if eqpt_color:
        options["color"] = eqpt_color
        equilibria = find_zeros(field, x_range, y_range)
        p += points(equilibria, size=50, **options)

    # Plot trajectories, if any initial conditions were given
    options["color"] = trajectory_color
    for init_cond, t_range in init_conds.items():
        solution = desolve_odeint(field, init_cond, t_range, [x, y])
        p += list_plot(solution, plotjoined=True, **options)

    return p

def axes3d(x_range, y_range, z_range, **options):
    """Draw x, y, and z axes in 3D, and optionally a grid on the xy-plane

    'x_range' should have the form (x, xmin, xmax), and likewise for 'y_range'
    and 'z_range'. Here the 'x' can be either a symbolic variable, or a string.
    Either way, it is only used as a label for the corresponding axis.

    Options:
        'grid' - Whether or not to draw a grid on the xy-plane (default True)
    All other options are as for the 'line3d' and 'arrow3d' commands, and are
    passed directly on to them.
    """

    grid = options.pop("grid", True)
    gridoptions = options.copy()
    options.setdefault("color", "black")
    gridoptions.setdefault("color", "gray")
    x, xmin, xmax = x_range
    y, ymin, ymax = y_range
    z, zmin, zmax = z_range
    xmin, xmax = min(0, xmin), max(0, xmax)
    ymin, ymax = min(0, ymin), max(0, ymax)
    zmin, zmax = min(0, zmin), max(0, zmax)
    p = Graphics()
    p += arrow3d((xmin, 0, 0), (xmax, 0, 0), width=1, **options)
    p += arrow3d((0, ymin, 0), (0, ymax, 0), width=1, **options)
    p += arrow3d((0, 0, zmin), (0, 0, zmax), width=1, **options)
    p += text3d(str(x), (xmax + (xmax - xmin)*0.05, 0, 0), **options)
    p += text3d(str(y), (0, ymax + (ymax - ymin)*0.05, 0), **options)
    p += text3d(str(z), (0, 0, zmax + (zmax - zmin)*0.05), **options)
    xgridsize = 10**int(log(1.2*max(-xmin, xmax), 10))
    ygridsize = 10**int(log(1.2*max(-ymin, ymax), 10))
    zgridsize = 10**int(log(1.2*max(-zmin, zmax), 10))
    gridmin = ceil(xmin/xgridsize) * xgridsize
    for xval in srange(gridmin, xmax+xgridsize/100.0, xgridsize):
        #p += text3d("{:.0f}".format(float(xval)), (xval, ygridsize/5.0, 0), **options)
        if grid:
            p += line3d(((xval, ymin, 0), (xval, ymax, 0)), thickness=1, **gridoptions)
    gridmin = ceil(ymin/ygridsize) * ygridsize
    for yval in srange(gridmin, ymax+ygridsize/100.0, ygridsize):
        #p += text3d("{:.0f}".format(float(yval)), (xgridsize/5.0, yval, 0), **options)
        if grid:
            p += line3d(((xmin, yval, 0), (xmax, yval, 0)), thickness=1, **gridoptions)
    gridmin = ceil(zmin/zgridsize) * zgridsize
    #for zval in srange(gridmin, zmax+zgridsize/100.0, zgridsize):
        #p += text3d("{:.0f}".format(float(zval)), (xgridsize/5.0, 0, zval), **options)
    return p

︡879b90f9-e08d-4bbd-9621-3c627f0fbe40︡{"done":true}
︠3eb0e9dc-002e-4289-a733-f3330c92e1afi︠
%md
### What sorts of long-term behavior can you <br />possibly get from a single (1D) <br />differential equation?
︡d21a4e52-55ad-49cc-a531-fb88b9d8e48b︡{"done":true,"md":"### What sorts of long-term behavior can you <br />possibly get from a single (1D) <br />differential equation?"}
︠354812a6-0ad7-4058-b27c-bbfeec26e37d︠




















︡cf1669ea-211a-45a1-830b-ab932da9e476︡
︠0ed77636-ddea-443b-91d0-426e05aaea32i︠
%md
### What sorts of long-term behavior can you <br />possibly get from a single (1D) <br />differential equation?

#### *Answer:*
- Constant solution (start on an equilibrium point, stay there forever)
- Approach a stable equilibrium point
- Grow without bound (increase to $\infty$ or decrease to $-\infty$)
︡c1bec377-217e-4faf-b9b0-366347eae817︡{"done":true,"md":"### What sorts of long-term behavior can you <br />possibly get from a single (1D) <br />differential equation?\n\n#### *Answer:*\n- Constant solution (start on an equilibrium point, stay there forever)\n- Approach a stable equilibrium point\n- Grow without bound (increase to $\\infty$ or decrease to $-\\infty$)"}
︠d0313197-ccd5-44f0-8abf-0b524fbb2cb4︠




















︡0aee81f3-d026-4dd7-943a-043806979e31︡
︠6eada0ca-9369-4f11-8ff5-1bc254c0fe09i︠
%md
### What sorts of long-term behavior can you <br />possibly have in a 2D system of <br />differential equations?
︡b927539f-bbe3-405a-8837-80ef378c069f︡{"done":true,"md":"### What sorts of long-term behavior can you <br />possibly have in a 2D system of <br />differential equations?"}
︠d3589a6a-8f94-40c3-a4a8-b6c7de53be96s︠
state_vars = list(var("X, Y"))
system = (
    4 - Y,
    2 - X,
)
field(X, Y) = system
t_range = srange(0, 10, 0.1)
ics = {
    (2, 4): t_range,
}
p1 = plot_phase_portrait(field, (X, -2, 4), (Y, -1, 5), ics,
                         nullcline_colors=(None, None), aspect_ratio=1)
p = p1 + text("Constant solution (single point trajectory)", (1, 6), fontsize=12)
p.show(axes_labels=("$X$", "$Y$"))
︡71b9dc91-af47-484d-8173-9d278f11aff8︡{"file":{"filename":"/home/user/.sage/temp/project-cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/349/tmp_tjnSO_.svg","show":true,"text":null,"uuid":"66781e82-e0ba-41f1-bafd-265224ddbd8c"},"once":false}︡{"done":true}︡
︠a4898b83-ab2c-447f-8bc0-f40ad9c143e9s︠
state_vars = list(var("X, Y"))
system = (
    -X + 0.4*Y,
    -2*Y - X,
)
field(X, Y) = system
t_range = srange(0, 100, 0.1)
ics = {
    (2, 2): t_range,
}
p2 = plot_phase_portrait(field, (X, -3, 3), (Y, -3, 3), ics,
                         nullcline_colors=(None, None), aspect_ratio=1)
p = p2 + text("Trajectory approaches stable eq. point", (0, 4), fontsize=12)
p.show(axes_labels=("$X$", "$Y$"))
︡148aec5c-d56f-402f-993b-ec446f97102c︡{"file":{"filename":"/home/user/.sage/temp/project-cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/349/tmp_5Ov6_O.svg","show":true,"text":null,"uuid":"ba3ec159-c583-41b9-88e8-961830df5fc4"},"once":false}︡{"done":true}︡
︠adf10e34-ad81-4651-9214-527740adc359s︠
state_vars = list(var("X, Y"))
system = (
    X - 0.5*Y,
    2*Y + X,
)
field(X, Y) = system
t_range = srange(0, 15.55, 0.1)
ics = {
    (0.1, -0.1): t_range,
}
p3 = plot_phase_portrait(field, (X, -1e9, 1e9), (Y, -1e9, 1e9), ics,
                         nullcline_colors=(None, None), aspect_ratio=1)
p = p3 + text('Trajectory is unbounded ("goes to infinity")', (0, 1.35E9), fontsize=12)
p.show(axes_labels=("$X$", "$Y$"))
︡fdf645aa-ffad-4acb-9726-f5141e89f0cb︡{"file":{"filename":"/home/user/.sage/temp/project-cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/349/tmp_KczdJa.svg","show":true,"text":null,"uuid":"725f96ea-6169-40f8-9da9-d5c14cb26e4c"},"once":false}︡{"done":true}︡
︠62840f28-7864-45c4-b3db-eb1d582c6defs︠
state_vars = list(var("X, Y"))
system = (
    Y,
    -X - (Y^3 - 0.1*Y),
)
field(X, Y) = system
t_range = srange(0, 10, 0.1)
ics = {
    (0.365, 0): t_range,
}
p4 = plot_phase_portrait(field, (X, -1, 1), (Y, -1, 1), ics,
                         nullcline_colors=(None, None), aspect_ratio=1)
p = p4 + text("Trajectory is periodic (closed loop orbit)", (0, 1.35), fontsize=12)
p.show(axes_labels=("$X$", "$Y$"))
︡90d65a3c-c377-465f-9a41-de5bf6fd6cf0︡{"file":{"filename":"/home/user/.sage/temp/project-cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/349/tmp_l04gmh.svg","show":true,"text":null,"uuid":"d740e69c-38d0-4b23-9612-42719b91429f"},"once":false}︡{"done":true}︡
︠3c817f4b-5b15-4fa7-876d-1b945704ab60s︠
state_vars = list(var("X, Y"))
system = (
    Y,
    -X - (Y^3 - 0.1*Y),
)
field(X, Y) = system
t_range = srange(0, 100, 0.1)
ics = {
    (1, -0.75): t_range,
}
p5 = plot_phase_portrait(field, (X, -1, 1), (Y, -1, 1), ics,
                         nullcline_colors=(None, None), aspect_ratio=1)
p = p5 + text("Trajectory approaches a periodic cycle (limit cycle attractor)", (0, 1.35), fontsize=12)
p.show(axes_labels=("$X$", "$Y$"))
︡3112f57b-e28e-47ee-ac2c-6eb19c4151f1︡{"file":{"filename":"/home/user/.sage/temp/project-cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/349/tmp_hTf9mt.svg","show":true,"text":null,"uuid":"46c339df-5ab1-4a5f-ba4b-32830fcc0320"},"once":false}︡{"done":true}︡
︠7ce98967-4cdf-4447-ba27-3e98425671fbs︠
state_vars = list(var("X, Y"))
system = (
    (Y - 0.25*X)*(X^2 - 1),
    X*(1 - Y^2),
)
field(X, Y) = system
t_range = srange(0, 100, 0.1)
ics = {
    (0.02, 0.02): t_range,
}
p6 = plot_phase_portrait(field, (X, -2, 2), (Y, -2, 2), ics,
                         aspect_ratio=1)
p = p6 + text('Multiple saddle points form a "polygon" of other', (0, 2.9), fontsize=12)
p += text('trajectories, and this trajectory is trapped inside!', (0, 2.7), fontsize=12)
p.show(axes_labels=("$X$", "$Y$"))
︡c20821bc-dde2-4e11-ac47-2ce5d374100c︡{"file":{"filename":"/home/user/.sage/temp/project-cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/349/tmp_S2ibxu.svg","show":true,"text":null,"uuid":"fbb28b5e-a493-4fa6-83d1-36092e73e9b9"},"once":false}︡{"done":true}︡
︠4e2e70f9-3b97-418b-9928-7fd6036ce963i︠
%md
### These are all of the long-term behaviors <br />that can happen to a trajectory in 2D!

#### (This fact is called the **Poincaré–Bendixson Theorem**.)
︡ab7d94a4-2f0c-4e96-8fa2-dd98bd8a4ee6︡︡{"done":true,"md":"### These are all of the long-term behaviors <br />that can happen to a trajectory in 2D!\n\n#### (This fact is called the **Poincaré–Bendixson Theorem**.)"}
︠92b88ef5-1a04-4ada-872f-86dec66dd4eas︠
ga = graphics_array(((p1, p2, p3), (p4, p5, p6)))
ga.show()
︡50654cbd-51ce-4776-b2de-3343decf21b3︡{"file":{"filename":"/projects/cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/.sage/temp/compute6-us/12567/tmp_HVzIHU.svg","show":true,"text":null,"uuid":"ddb1b0ed-0614-4ac3-9739-6384719feb27"},"once":false}︡{"done":true}︡
︠adbdcfff-b86d-4ad8-8f80-517a58a5b730︠




















︡8b906a5c-a215-41dc-a568-c53f94fa3a31︡
︠9f662b81-6002-4734-8ac5-5b92e6f93962i︠
%md
### In 3D (and higher dimensions), there's a <br />new kind of behavior that does not fit <br />any of these patterns...
︡47de8462-297f-4b7c-b5e9-913559f79a81︡︡{"done":true,"md":"### In 3D (and higher dimensions), there's a <br />new kind of behavior that does not fit <br />any of these patterns..."}
︠531def94-def7-4a29-820e-374f6227f6f2s︠
state_vars = list(var("X, Y, Z"))
system = (
    X*(1 - X) - 5/3*X/(1/3 + X)*Y,
    5/3*X/(1/3 + X)*Y - 0.4*Y - 0.05*Y/(0.5 + Y)*Z,
    0.05*Y/(0.5 + Y)*Z - 0.01*Z,
)
field(X, Y, Z) = system
t_range = srange(0, 10000, 0.1)
solution = desolve_odeint(field, (1, 1, 1), t_range, state_vars)
start = 6000/0.1
#p = axes3d((X, 0, 1), (Y, 0, 0.7), (Z, 0, 12), grid=False)
p = Graphics()
p += line3d(((0,0,7), (1,  0, 7)), color="black") + text3d("X", (1.1,0,7), color="black")
p += line3d(((0,0,7), (0,0.5, 7)), color="black") + text3d("Y", (0,0.6,7), color="black")
p += line3d(((0,0,7), (0,  0,11)), color="black") + text3d("Z", (0,0, 11.1), color="black")
p += list_plot(solution[start::10,:], plotjoined=True, color="red")
p.show(frame=False, aspect_ratio=(3,6,1))
︡17ddb4b1-dba7-4fc1-9523-c949924a6735︡{"file":{"filename":"6c6657bf-a617-4c56-b9df-b14ceeb853a2.sage3d","uuid":"6c6657bf-a617-4c56-b9df-b14ceeb853a2"}}︡{"done":true}︡
︠4064dcf0-9d47-4908-b727-4a76e58a8868s︠
start = int(9000/0.1)
p  = list_plot(zip(t_range[start:], solution[start:,0]), plotjoined=True, color="red",   legend_label="$X$")
p += list_plot(zip(t_range[start:], solution[start:,1]), plotjoined=True, color="green", legend_label="$Y$")
p += list_plot(zip(t_range[start:], solution[start:,2]), plotjoined=True, color="blue",  legend_label="$Z$")
show(p, ymin=0, figsize=5, axes_labels=("$t$", "Populations"))
︡3354cbc9-aa13-4e96-9d62-664f8b238a8f︡{"file":{"filename":"/projects/cdb600df-721d-4ab1-aa2a-b8fb0f7750b8/.sage/temp/compute6-us/12567/tmp__KzqXD.svg","show":true,"text":null,"uuid":"655bf74e-1ff8-46bf-a2e7-7d27c1154050"},"once":false}︡{"done":true}︡
︠06247c23-3d71-4951-bcb3-a5da014c436ei︠
%md
### It never precisely repeats itself, so it's not oscillating. But it also does not go to equilibrium, nor does it go to infinity.

### This is called *chaotic* behavior, and this will be the topic of Chapter 5.
︡81e08f4b-4ca9-4bf6-b5c9-aefa0f7acdd7︡{"done":true,"md":"### It never precisely repeats itself, so it's not oscillating. But it also does not go to equilibrium, nor does it go to infinity. \n\n### This is called *chaotic* behavior, and this will be the topic of Chapter 5."}
︠42d725ca-ac69-4764-bc0c-cf0ba7ec911d︠









